"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema.validators import ValidationError
from cfnlint.rules.resources.stepfunctions.StateMachineDefinition import (
    StateMachineDefinition,
)


@pytest.fixture(scope="module")
def rule():
    rule = StateMachineDefinition()
    yield rule


# These state machines for testing have long lines
# with lots of indentation
# ruff: noqa: E501


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid string json",
            {
                "Comment": (
                    "An example of the Amazon States "
                    "Language for notification on an "
                    "AWS Batch job completion"
                ),
                "StartAt": "Submit Batch Job",
                "TimeoutSeconds": 3600,
                "States": {
                    "Submit Batch Job": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::batch:submitJob.sync",
                        "Parameters": {
                            "JobName": "BatchJobNotification",
                            "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                            "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                        },
                        "Next": "Notify Success",
                        "Catch": [
                            {"ErrorEquals": ["States.ALL"], "Next": "Notify Failure"}
                        ],
                    },
                    "Notify Success": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::sns:publish",
                        "Parameters": {
                            "Message": "Batch job submitted through Step Functions succeeded",
                            "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                        },
                        "End": True,
                    },
                    "Notify Failure": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::sns:publish",
                        "Parameters": {
                            "Message": "Batch job submitted through Step Functions failed",
                            "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                        },
                        "End": True,
                    },
                },
            },
            [],
        ),
        (
            "Fan out example",
            {
                "Comment": "An example of the Amazon States Language for fanning out AWS Batch job",
                "StartAt": "Generate batch job input",
                "TimeoutSeconds": 3600,
                "States": {
                    "Generate batch job input": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::lambda:invoke",
                        "OutputPath": "$.Payload",
                        "Parameters": {
                            "FunctionName": "<GENERATE_BATCH_JOB_INPUT_LAMBDA_FUNCTION_NAME>"
                        },
                        "Next": "Fan out batch jobs",
                    },
                    "Fan out batch jobs": {
                        "Comment": "Start multiple executions of batch job depending on pre-processed data",
                        "Type": "Map",
                        "End": True,
                        "ItemsPath": "$",
                        "Parameters": {"BatchNumber.$": "$$.Map.Item.Value"},
                        "Iterator": {
                            "StartAt": "Submit Batch Job",
                            "States": {
                                "Submit Batch Job": {
                                    "Type": "Task",
                                    "Resource": "arn:aws:states:::batch:submitJob.sync",
                                    "Parameters": {
                                        "JobName": "BatchJobFanOut",
                                        "JobQueue": "<BATCH_QUEUE_ARN>",
                                        "JobDefinition": "<BATCH_JOB_DEFINITION_ARN>",
                                    },
                                    "End": True,
                                }
                            },
                        },
                    },
                },
            },
            [],
        ),
        (
            "Error handling",
            """
            {
                "Comment": "An example of the Amazon States Language for notification on an AWS Batch job completion",
                "StartAt": "Submit Batch Job",
                "TimeoutSeconds": 3600,
                "States": {
                    "Submit Batch Job": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::batch:submitJob.sync",
                        "Parameters": {
                            "JobName": "BatchJobNotification",
                            "JobQueue": "arn:aws:batch:us-west-2:123456789012:job-queue/BatchJobQueue-123456789abcdef",
                            "JobDefinition": "arn:aws:batch:us-west-2:123456789012:job-definition/BatchJobDefinition-123456789abcdef:1",
                        },
                        "Next": "Notify Success",
                        "Retry": [
                            {
                                "ErrorEquals": ["States.ALL"],
                                "IntervalSeconds": 30,
                                "MaxAttempts": 2,
                                "BackoffRate": 1.5,
                            }
                        ],
                        "Catch": [
                            {"ErrorEquals": ["States.ALL"], "Next": "Notify Failure"}
                        ],
                    },
                    "Notify Success": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::sns:publish",
                        "Parameters": {
                            "Message": "Batch job submitted through Step Functions succeeded",
                            "TopicArn": "arn:aws:sns:us-west-2:123456789012:StepFunctionsSample-BatchJobManagement12345678-9abc-def0-1234-567890abcdef-SNSTopic-A2B3C4D5E6F7G",
                        },
                        "End": True,
                    },
                    "Notify Failure": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::sns:publish",
                        "Parameters": {
                            "Message": "Batch job submitted through Step Functions failed",
                            "TopicArn": "arn:aws:sns:us-west-2:123456789012:StepFunctionsSample-BatchJobManagement12345678-9abc-def0-1234-567890abcdef-SNSTopic-A2B3C4D5E6F7G",
                        },
                        "End": True,
                    },
                },
            }
            """,
            [],
        ),
        (
            "Transfer data records",
            {
                "Comment": "An example of the Amazon States Language for reading messages from a DynamoDB table and sending them to SQS",
                "StartAt": "Seed the DynamoDB Table",
                "TimeoutSeconds": 3600,
                "States": {
                    "Seed the DynamoDB Table": {
                        "Type": "Task",
                        "Resource": "arn:aws:lambda:us-east-1:123456789012:function:sqsconnector-SeedingFunction-T3U43VYDU5OQ",
                        "ResultPath": "$.List",
                        "Next": "For Loop Condition",
                    },
                    "For Loop Condition": {
                        "Type": "Choice",
                        "Choices": [
                            {
                                "Not": {
                                    "Variable": "$.List[0]",
                                    "StringEquals": "DONE",
                                },
                                "Next": "Read Next Message from DynamoDB",
                            }
                        ],
                        "Default": "Succeed",
                    },
                    "Read Next Message from DynamoDB": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::dynamodb:getItem",
                        "Parameters": {
                            "TableName": "sqsconnector-DDBTable-1CAFOJWP8QD6I",
                            "Key": {"MessageId": {"S.$": "$.List[0]"}},
                        },
                        "ResultPath": "$.DynamoDB",
                        "Next": "Send Message to SQS",
                    },
                    "Send Message to SQS": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::sqs:sendMessage",
                        "Parameters": {
                            "MessageBody.$": "$.DynamoDB.Item.Message.S",
                            "QueueUrl": "https://sqs.us-east-1.amazonaws.com/123456789012/sqsconnector-SQSQueue-QVGQBW134PWK",
                        },
                        "ResultPath": "$.SQS",
                        "Next": "Pop Element from List",
                    },
                    "Pop Element from List": {
                        "Type": "Pass",
                        "Parameters": {"List.$": "$.List[1:]"},
                        "Next": "For Loop Condition",
                    },
                    "Succeed": {"Type": "Succeed"},
                },
            },
            [],
        ),
        (
            "A map",
            {
                "Comment": "An example of the Amazon States Language for reading messages from an SQS queue and iteratively processing each message.",
                "StartAt": "Read messages from SQS Queue",
                "States": {
                    "Read messages from SQS Queue": {
                        "Type": "Task",
                        "Resource": "arn:aws:states:::lambda:invoke",
                        "OutputPath": "$.Payload",
                        "Parameters": {
                            "FunctionName": "MapSampleProj-ReadFromSQSQueueLambda-1MY3M63RMJVA9"
                        },
                        "Next": "Are there messages to process?",
                    },
                    "Are there messages to process?": {
                        "Type": "Choice",
                        "Choices": [
                            {
                                "Variable": "$",
                                "StringEquals": "No messages",
                                "Next": "Finish",
                            }
                        ],
                        "Default": "Process messages",
                    },
                    "Process messages": {
                        "Type": "Map",
                        "Next": "Finish",
                        "ItemsPath": "$",
                        "Parameters": {
                            "MessageNumber.$": "$$.Map.Item.Index",
                            "MessageDetails.$": "$$.Map.Item.Value",
                        },
                        "Iterator": {
                            "StartAt": "Write message to DynamoDB",
                            "States": {
                                "Write message to DynamoDB": {
                                    "Type": "Task",
                                    "Resource": "arn:aws:states:::dynamodb:putItem",
                                    "ResultPath": None,
                                    "Parameters": {
                                        "TableName": "MapSampleProj-DDBTable-YJDJ1MKIN6C5",
                                        "ReturnConsumedCapacity": "TOTAL",
                                        "Item": {
                                            "MessageId": {
                                                "S.$": "$.MessageDetails.MessageId"
                                            },
                                            "Body": {"S.$": "$.MessageDetails.Body"},
                                        },
                                    },
                                    "Next": "Remove message from SQS queue",
                                },
                                "Remove message from SQS queue": {
                                    "Type": "Task",
                                    "Resource": "arn:aws:states:::lambda:invoke",
                                    "InputPath": "$.MessageDetails",
                                    "ResultPath": None,
                                    "Parameters": {
                                        "FunctionName": "MapSampleProj-DeleteFromSQSQueueLambda-198J2839ZO5K2",
                                        "Payload": {
                                            "ReceiptHandle.$": "$.ReceiptHandle"
                                        },
                                    },
                                    "Next": "Publish message to SNS topic",
                                },
                                "Publish message to SNS topic": {
                                    "Type": "Task",
                                    "Resource": "arn:aws:states:::sns:publish",
                                    "InputPath": "$.MessageDetails",
                                    "Parameters": {
                                        "Subject": "Message from Step Functions!",
                                        "Message.$": "$.Body",
                                        "TopicArn": "arn:aws:sns:us-east-1:012345678910:MapSampleProj-SNSTopic-1CQO4HQ3IR1KN",
                                    },
                                    "End": True,
                                },
                            },
                        },
                    },
                    "Finish": {"Type": "Succeed"},
                },
            },
            [],
        ),
        (
            "Test functions",
            {
                "StartAt": "Task1",
                "States": {
                    "Task1": {
                        "Type": "Task",
                        "Credentials": {
                            "Data": "input data",
                            "Algorithm": "SHA-1",
                        },
                        "Resource": "arn:aws:states:::batch:submitJob.sync",
                        "End": True,
                    },
                },
            },
            [],
        ),
        (
            "Choices",
            {
                "StartAt": "ChoiceStateX",
                "States": {
                    "ChoiceStateX": {
                        "Type": "Choice",
                        "Choices": [
                            {
                                "Not": {
                                    "Variable": "$.type",
                                    "StringEquals": "Private",
                                },
                                "Next": "Public",
                            },
                            {
                                "Variable": "$.value",
                                "NumericEquals": 0,
                                "Next": "ValueIsZero",
                            },
                            {
                                "And": [
                                    {
                                        "Variable": "$.value",
                                        "NumericGreaterThanEquals": 20,
                                    },
                                    {"Variable": "$.value", "NumericLessThan": 30},
                                ],
                                "Next": "ValueInTwenties",
                            },
                            {
                                "Variable": "$.value",
                                "NumericLessThan": 0,
                                "Next": "ValueIsNegative",
                            },
                        ],
                        "Default": "DefaultState",
                    },
                    "Public": {
                        "Type": "Task",
                        "Resource": "arn:aws:lambda:us-east-1:123456789012:function:Foo",
                        "Next": "NextState",
                    },
                    "ValueIsZero": {
                        "Type": "Task",
                        "Resource": "arn:aws:lambda:us-east-1:123456789012:function:Zero",
                        "Next": "NextState",
                    },
                    "ValueInTwenties": {
                        "Type": "Task",
                        "Resource": "arn:aws:lambda:us-east-1:123456789012:function:Bar",
                        "Next": "NextState",
                    },
                    "ValueIsNegative": {
                        "Type": "Succeed",
                    },
                    "DefaultState": {"Type": "Fail", "Cause": "No Matches!"},
                },
            },
            [],
        ),
        (
            "Invalid configuration",
            {
                "StartAt": "SampleChoice",
                "States": {
                    "NoType": {},
                    "SampleChoice": {
                        "Default": "__SucceedEntryPoint__",
                        "Type": "Choice",
                    },
                    "__SucceedEntryPoint__": {"Type": "Pass"},
                },
            },
            [
                ValidationError(
                    "'Type' is a required property",
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,128}$",
                            "required",
                        ]
                    ),
                    path=deque(["States", "NoType"]),
                ),
                ValidationError(
                    "'Choices' is a required property",
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,128}$",
                            "allOf",
                            0,
                            "then",
                            "required",
                        ]
                    ),
                    path=deque(["States", "SampleChoice"]),
                ),
                ValidationError(
                    "Only one of ['Next', 'End'] is a required property",
                    rule=StateMachineDefinition(),
                    validator="requiredXor",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,128}$",
                            "allOf",
                            3,
                            "then",
                            "requiredXor",
                        ]
                    ),
                    path=deque(["States", "__SucceedEntryPoint__"]),
                ),
            ],
        ),
        (
            "Invalid string definition",
            """
            {
                "States": {
                    "NoType": {}
                }
            }
            """,
            [
                ValidationError(
                    "'Type' is a required property at 'States/NoType'",
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,128}$",
                            "required",
                        ]
                    ),
                    path=deque(["States", "NoType"]),
                ),
                ValidationError(
                    "'StartAt' is a required property",
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "required",
                        ]
                    ),
                    path=deque([]),
                ),
            ],
        ),
    ],
)
def test_validate(
    name,
    instance,
    expected,
    rule,
    validator,
):
    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == expected, f"{name!r} test failed with {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import os
from test.integration import BaseCliTestCase

from jsonpatch import InvalidJsonPatch, JsonPatch


class TestPatches(BaseCliTestCase):
    """Test Patches"""

    def test_patches(self):
        """Test ignoring certain rules"""
        for root, _, files in os.walk(
            "src/cfnlint/data/ExtendedProviderSchemas", topdown=True
        ):
            for name in files:
                if name.endswith(".json"):
                    with open(os.path.join(root, name)) as fh:
                        patches = json.load(fh)
                        try:
                            JsonPatch(patches)
                        except InvalidJsonPatch:
                            raise Exception(f"Invalid patch: {name}")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
from collections import deque
from typing import Any, Iterator

import regex as re

from cfnlint.helpers import AVAILABILITY_ZONES, REGEX_SUB_PARAMETERS
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.jsonschema._typing import ResolutionResult


def unresolvable(validator: Validator, instance: Any) -> ResolutionResult:
    return
    yield


def ref(validator: Validator, instance: Any) -> ResolutionResult:
    if not isinstance(instance, (str, dict)):
        return

    for instance, instance_validator, _ in validator.resolve_value(instance):
        if validator.is_type(instance, "string"):
            # if the ref is to pseudo-parameter or parameter we can validate the values
            for v, c in instance_validator.context.ref_value(instance):
                yield v, instance_validator.evolve(context=c), None
            return


def find_in_map(validator: Validator, instance: Any) -> ResolutionResult:
    if not validator.is_type(instance, "array"):
        return
    if len(instance) not in [3, 4]:
        return

    default_value_found = None
    if len(instance) == 4:
        options = instance[3]
        if validator.is_type(options, "object"):
            if "DefaultValue" in options:
                for value, v, _ in validator.resolve_value(options["DefaultValue"]):
                    yield value, v.evolve(
                        context=v.context.evolve(
                            path=v.context.path.evolve(
                                value_path=deque([4, "DefaultValue"])
                            )
                        ),
                    ), None
                default_value_found = True

    if not default_value_found and not validator.context.mappings.maps:
        if validator.context.mappings.is_transform:
            return
        yield None, validator, ValidationError(
            (
                f"{instance[0]!r} is not one of "
                f"{list(validator.context.mappings.maps.keys())!r}"
            ),
            path=deque([0]),
        )

    if (
        validator.is_type(instance[0], "string")
        and (
            validator.is_type(instance[1], "string")
            or validator.is_type(instance[1], "integer")
        )
        and validator.is_type(instance[2], "string")
    ):
        map = validator.context.mappings.maps.get(instance[0])
        if map is None:
            if not default_value_found:
                yield None, validator, ValidationError(
                    (
                        f"{instance[0]!r} is not one of "
                        f"{list(validator.context.mappings.maps.keys())!r}"
                    ),
                    path=deque([0]),
                )
            return

        top_key = map.keys.get(instance[1])
        if top_key is None:
            if map.is_transform:
                return
            if not default_value_found:
                yield None, validator, ValidationError(
                    (
                        f"{instance[1]!r} is not one of "
                        f"{list(map.keys.keys())!r} for "
                        f"mapping {instance[0]!r}"
                    ),
                    path=deque([1]),
                )
            return

        value = top_key.keys.get(instance[2])
        if value is None:
            if top_key.is_transform:
                return
            if not default_value_found:
                yield value, validator, ValidationError(
                    (
                        f"{instance[2]!r} is not one of "
                        f"{list(top_key.keys.keys())!r} for mapping "
                        f"{instance[0]!r} and key {instance[1]!r}"
                    ),
                    path=deque([2]),
                )


def get_azs(validator: Validator, instance: Any) -> ResolutionResult:
    if not isinstance(instance, (str, dict)):
        return

    for instance, v, _ in validator.resolve_value(instance):
        if v.is_type(instance, "string"):
            if instance == "":
                for region in v.context.regions:
                    yield (
                        AVAILABILITY_ZONES.get(region),
                        v,
                        None,
                    )
            # if the ref is to pseudo-parameter or parameter we can validate the values
            elif instance in AVAILABILITY_ZONES:
                yield AVAILABILITY_ZONES.get(instance), v, None


def _join_expansion(validator: Validator, instances: Any) -> Iterator[Any]:
    if len(instances) == 0:
        return

    if len(instances) == 1:
        for value, _, _ in validator.resolve_value(instances[0]):
            if not isinstance(value, (str, int, float, bool)):
                raise ValueError(f"Incorrect value type for {value!r}")
            yield [value]
        return

    for value, _, _ in validator.resolve_value(instances[0]):
        if not isinstance(value, (str, int, float, bool)):
            raise ValueError(f"Incorrect value type for {value!r}")
        for values in _join_expansion(validator, instances[1:]):
            yield [value] + values


def join(validator: Validator, instance: Any) -> ResolutionResult:
    # quick validations
    if not validator.is_type(instance, "array"):
        return
    if not len(instance) == 2:
        return

    for delimiter, delimiter_v, _ in validator.resolve_value(instance[0]):
        if not delimiter_v.is_type(delimiter, "string"):
            continue
        for values, values_v, _ in validator.resolve_value(instance[1]):
            if not values_v.is_type(values, "array"):
                continue
            try:
                for value in _join_expansion(values_v, values):
                    yield delimiter.join(value), values_v, None
            except ValueError:
                return


def select(validator: Validator, instance: Any) -> ResolutionResult:
    # quick validations
    if not validator.is_type(instance, "array"):
        return
    if not len(instance) == 2:
        return

    # get the values from the list
    indexes = validator.resolve_value(instance[0])
    objs = validator.resolve_value(instance[1])

    for i, _, _ in indexes:
        for obj, obj_v, _ in objs:
            try:
                i = int(i)
            except ValueError:
                continue
            if not validator.is_type(obj, "array"):
                continue
            if len(obj) <= i:
                continue
            yield from obj_v.resolve_value(obj[i])


def split(validator: Validator, instance: Any) -> ResolutionResult:
    if not validator.is_type(instance, "array"):
        return
    if not len(instance) == 2:
        return

    for delimiter, _, _ in validator.resolve_value(instance[0]):
        for source_string, source_v, _ in validator.resolve_value(instance[1]):
            if not source_v.is_type(delimiter, "string"):
                continue
            if not source_v.is_type(source_string, "string"):
                continue

            yield source_string.split(delimiter), source_v, None


def _sub_parameter_expansion(
    validator: Validator, parameters: dict[str, Any]
) -> Iterator[dict[str, Any]]:
    parameters = parameters.copy()
    if len(parameters) == 0:
        yield {}
        return

    if len(parameters) == 1:
        for key, value in parameters.items():
            for resolved_value, _, _ in validator.resolve_value(value):
                yield {key: resolved_value}
        return

    key = list(parameters.keys())[0]
    value = parameters.pop(key)
    for resolved_value, _, _ in validator.resolve_value(value):
        for values in _sub_parameter_expansion(validator, parameters):
            yield dict({key: resolved_value}, **values)


def _sub_string(validator: Validator, string: str) -> ResolutionResult:
    sub_regex = re.compile(r"(\${([^!].*?)})")

    def _replace(matchobj):
        nonlocal validator
        for value, c in validator.context.ref_value(matchobj.group(2).strip()):
            if not isinstance(value, (str, int, float, bool)):
                raise ValueError(f"Parameter {matchobj.group(2)!r} has wrong type")

            validator = validator.evolve(
                context=validator.context.evolve(
                    ref_values=c.ref_values,
                )
            )
            return str(value)
        raise ValueError(f"No matches for {matchobj.group(2)!r}")

    try:
        yield re.sub(sub_regex, _replace, string), validator, None
    except ValueError:
        return


def sub(validator: Validator, instance: Any) -> ResolutionResult:
    if not (
        validator.is_type(instance, "array") or validator.is_type(instance, "string")
    ):
        return

    if validator.is_type(instance, "array"):
        if len(instance) != 2:
            return

        string = instance[0]
        parameters = instance[1]
        if not validator.is_type(string, "string"):
            return
        if not validator.is_type(parameters, "object"):
            return

        for resolved_parameters in _sub_parameter_expansion(validator, parameters):
            resolved_validator = validator.evolve(
                context=validator.context.evolve(
                    ref_values=resolved_parameters,
                )
            )
            yield from _sub_string(resolved_validator, string)

        return

    # its a string
    sub_parameters = REGEX_SUB_PARAMETERS.findall(instance)
    parameters = {}
    for parameter in sub_parameters:
        if "." in parameter:
            parameters[parameter] = {"Fn::GetAtt": parameter}
        else:
            parameters[parameter] = {"Ref": parameter}
    for resolved_parameters in _sub_parameter_expansion(validator, parameters):
        resolved_validator = validator.evolve(
            context=validator.context.evolve(
                ref_values=resolved_parameters,
            )
        )
        yield from _sub_string(resolved_validator, instance)
        # yield from _sub_string(validator, instance)


def if_(validator: Validator, instance: Any) -> ResolutionResult:
    if not validator.is_type(instance, "array"):
        return

    if len(instance) != 3:
        return

    for i in [1, 2]:
        for value, v, err in validator.resolve_value(instance[i]):
            yield (
                value,
                v.evolve(
                    context=v.context.evolve(
                        path=v.context.path.evolve(value_path=deque([i])),
                    ),
                ),
                err,
            )


def to_json_string(validator: Validator, instance: Any) -> ResolutionResult:
    for value, v, err in validator.resolve_value(instance):
        yield json.dumps(value), v, err


# not all functions need to be resolved.  These functions
# allow us to pull up values from nested functions
# allowing us to test the possible values against the schema
fn_resolvers: dict[str, Any] = {
    "Fn::Base64": unresolvable,
    "Fn::Cidr": unresolvable,
    "Fn::FindInMap": find_in_map,
    "Fn::ForEach": unresolvable,
    "Fn::GetAtt": unresolvable,
    "Fn::GetAZs": get_azs,
    "Fn::ImportValue": unresolvable,
    "Fn::If": if_,
    "Fn::Join": join,
    "Fn::Select": select,
    "Fn::Split": split,
    "Fn::Sub": sub,
    "Fn::Transform": unresolvable,
    "Fn::ToJsonString": to_json_string,
    "Fn::Equals": unresolvable,
    "Fn::Or": unresolvable,
    "Fn::And": unresolvable,
    "Fn::Not": unresolvable,
    "Condition": unresolvable,
    "Fn::Length": unresolvable,
    "Ref": ref,
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint._typing import RuleMatches
from cfnlint.helpers import REGEX_DYN_REF, REGEX_DYN_REF_SSM
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class Password(CloudFormationLintRule):
    """Check if Password Properties are properly configured"""

    id = "W2501"
    shortdesc = "Check if Password Properties are correctly configured"
    description = (
        "Password properties should not be strings and if parameter using NoEcho"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/best-practices.html#creds"
    tags = ["parameters", "passwords", "security", "dynamic reference"]

    def match(self, cfn: Template) -> RuleMatches:
        """Check CloudFormation Password Parameters"""

        matches = []
        password_properties = [
            "AccountPassword",
            "AdminPassword",
            "ADDomainJoinPassword",
            "CrossRealmTrustPrincipalPassword",
            "KdcAdminPassword",
            "Password",
            "DbPassword",
            "MasterUserPassword",
            "PasswordParam",
        ]

        parameters = list(cfn.template.get("Parameters", {}).keys())
        fix_params = []
        for password_property in password_properties:
            # Build the list of refs
            refs = cfn.search_deep_keys(password_property)
            trees = []
            for tree in refs:
                if len(tree) > 2:
                    if tree[0] == "Resources" and tree[2] == "Properties":
                        trees.append(tree)

            for tree in trees:
                obj = tree[-1]
                if isinstance(obj, (str)):
                    if re.match(REGEX_DYN_REF, obj):
                        if re.match(REGEX_DYN_REF_SSM, obj):
                            message = (
                                "Password should use a secure dynamic reference for"
                                f" {'/'.join(map(str, tree[:-1]))}"
                            )
                            matches.append(RuleMatch(tree[:-1], message))
                    else:
                        message = (
                            "Password shouldn't be hardcoded for"
                            f" {'/'.join(map(str, tree[:-1]))}"
                        )
                        matches.append(RuleMatch(tree[:-1], message))
                elif isinstance(obj, dict):
                    if len(obj) == 1:
                        for key, value in obj.items():
                            if key == "Ref":
                                if value in parameters:
                                    param = cfn.template["Parameters"][value]
                                    if "NoEcho" in param:
                                        if not param["NoEcho"]:
                                            fix_params.append(
                                                {
                                                    "Name": value,
                                                    "Use": password_property,
                                                }
                                            )
                                    else:
                                        fix_params.append(
                                            {"Name": value, "Use": password_property}
                                        )
                    else:
                        message = (
                            "Inappropriate map found for password on"
                            f" {'/'.join(map(str, tree[:-1]))}"
                        )
                        matches.append(RuleMatch(tree[:-1], message))

        for paramname in fix_params:
            message = (
                f'Parameter {paramname["Name"]} used as {paramname["Use"]}, therefore'
                " NoEcho should be True"
            )
            tree = ["Parameters", paramname["Name"]]
            matches.append(RuleMatch(tree, message))
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Any

import regex as re

from cfnlint.helpers import REGEX_SUB_PARAMETERS
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules import CloudFormationLintRule


class SubParametersUsed(CloudFormationLintRule):
    """Check if Sub Parameters are used"""

    id = "W1019"
    shortdesc = "Validate that parameters to a Fn::Sub are used"
    description = "Validate that Fn::Sub Parameters are used"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub"]

    def validate(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        if validator.is_type(instance, "string"):
            return

        variables = re.findall(REGEX_SUB_PARAMETERS, instance[0])
        for v, _ in instance[1].items():
            if v not in variables:
                yield ValidationError(
                    f"Parameter {v!r} not used in 'Fn::Sub'", path=deque([1, v])
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class SnapStartSupported(CfnLintKeyword):
    """Check if Lambda function using SnapStart has the correct runtimes"""

    id = "E2530"
    shortdesc = "SnapStart supports the configured runtime"
    description = (
        "To properly leverage SnapStart, you must have a runtime of Java11 or greater"
    )
    source_url = "https://docs.aws.amazon.com/lambda/latest/dg/snapstart.html"
    tags = ["resources", "lambda"]

    def __init__(self):
        super().__init__(["Resources/AWS::Lambda::Function/Properties"])
        self.child_rules = {"I2530": None}
        self.regions = [
            "us-east-2",
            "us-east-1",
            "us-west-1",
            "us-west-2",
            "af-south-1",
            "ap-east-1",
            "ap-southeast-3",
            "ap-south-1",
            "ap-northeast-2",
            "ap-northeast-3",
            "ap-southeast-1",
            "ap-southeast-2",
            "ap-northeast-1",
            "ca-central-1",
            "eu-central-1",
            "eu-west-1",
            "eu-west-2",
            "eu-south-1",
            "eu-west-3",
            "eu-north-1",
            "me-south-1",
            "sa-east-1",
        ]

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for scenario in validator.cfn.get_object_without_conditions(
            instance,
            ["Runtime", "SnapStart"],
        ):
            props = scenario.get("Object")

            runtime = props.get("Runtime")
            snap_start = props.get("SnapStart")
            if not snap_start:
                if self.child_rules["I2530"]:
                    if all(
                        region in self.regions for region in validator.context.regions
                    ):
                        yield from self.child_rules["I2530"].validate(  # type: ignore
                            runtime,
                        )
                continue

            if snap_start.get("ApplyOn") != "PublishedVersions":
                continue

            if any(region not in self.regions for region in validator.context.regions):
                unsupported_regions = [
                    region
                    for region in validator.context.regions
                    if region not in self.regions
                ]
                yield ValidationError(
                    (
                        "'SnapStart' enabled functions are not supported in "
                        f"{unsupported_regions!r}"
                    ),
                    path=deque(["SnapStart", "ApplyOn"]),
                )

            # Validate runtime is a string before using startswith
            if not isinstance(runtime, str):
                continue

            if (
                runtime
                and (not runtime.startswith("java"))
                and runtime not in ["java8.al2", "java8"]
            ):
                yield ValidationError(
                    f"{runtime!r} is not supported for 'SnapStart' enabled functions",
                    path=deque(["SnapStart", "ApplyOn"]),
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class ApproachingMaxLength(CloudFormationLintRule):
    """Check maximum Parameter name size limit"""

    id = "I2011"
    shortdesc = "Parameter name limit"
    description = (
        "Check the size of Parameter names in the template is approaching the upper"
        " limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["parameters", "limits"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


@dataclass(frozen=True)
class Condition:
    instance: Any = field(init=True)
    status: bool | None = field(init=True, default=None)

    def evolve(self, status: bool | None) -> "Condition":
        cls = self.__class__

        if self.status is not None:
            if status != self.status:
                raise ValueError("Resetting status to condition")

        return cls(
            instance=self.instance,
            status=status,
        )


@dataclass(frozen=True)
class Conditions:
    # Template level condition management
    conditions: dict[str, Condition] = field(init=True, default_factory=dict)

    @classmethod
    def create_from_instance(cls, conditions: dict[str, Any]) -> "Conditions":
        obj = {}
        if not isinstance(conditions, dict):
            raise ValueError("Conditions must be a object")
        for k, v in conditions.items():
            try:
                obj[k] = Condition(v)
            except ValueError:
                pass

        return cls(conditions=obj)

    def evolve(self, status: dict[str, bool]) -> "Conditions":
        cls = self.__class__

        conditions: dict[str, Condition] = {}
        for condition, value in self.conditions.items():
            s = status.get(condition, value.status)
            try:
                conditions[condition] = value.evolve(status=s)
            except ValueError as e:
                raise ValueError(f"Error evolving condition {condition}") from e

        return cls(
            conditions=conditions,
        )

    @property
    def status(self) -> dict[str, bool]:
        obj = {}
        for name, c in self.conditions.items():
            if c.status is not None:
                obj[name] = c.status

        return obj



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path, create_context_for_template
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.parameters.NoEcho import NoEcho
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = NoEcho()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {
            "Parameters": {
                "Echo": {"Type": "String", "NoEcho": "true"},
                "EchoTrue": {
                    "Type": "String",
                    "NoEcho": True,
                },
                "NoEcho": {"Type": "String", "NoEcho": "false"},
                "MyParameter": {"Type": "String"},
            },
            "Resources": {},
        },
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "An Invalid Ref",
            {"Ref": []},
            deque(["Metadata"]),
            [],
        ),
        (
            "A ref to something that isn't a parameter",
            {"Ref": "AWS::Region"},
            deque(["Metadata"]),
            [],
        ),
        (
            "Using NoEcho in Resources Properties",
            {"Ref": "Echo"},
            deque(["Resources", "MyResource", "Properties", "Name"]),
            [],
        ),
        (
            "Using NoEcho in Resource Metadata",
            {"Ref": "Echo"},
            deque(["Resources", "MyResource", "Metadata"]),
            [
                ValidationError(
                    "Don't use 'NoEcho' parameter 'Echo' in resource metadata",
                    rule=NoEcho(),
                    path=deque(["Ref"]),
                )
            ],
        ),
        (
            "Using NoEcho in Metadata",
            {"Ref": "Echo"},
            deque(["Metadata"]),
            [
                ValidationError(
                    "Don't use 'NoEcho' parameter 'Echo' in 'Metadata'",
                    rule=NoEcho(),
                    path=deque(["Ref"]),
                )
            ],
        ),
        (
            "Using NoEcho in Metadata",
            {"Ref": "EchoTrue"},
            deque(["Metadata"]),
            [
                ValidationError(
                    "Don't use 'NoEcho' parameter 'EchoTrue' in 'Metadata'",
                    rule=NoEcho(),
                    path=deque(["Ref"]),
                )
            ],
        ),
        (
            "Using Non NoEcho in Metadata",
            {"Ref": "NoEcho"},
            deque(["Metadata"]),
            [],
        ),
        (
            "Using a parameter in Metadata",
            {"Ref": "MyParameter"},
            deque(["Metadata"]),
            [],
        ),
        (
            "Short list for path",
            {"Ref": "MyParameter"},
            deque([]),
            [],
        ),
        (
            "Using NoEcho in Outputs",
            {"Ref": "Echo"},
            deque(["Outputs", "Name", "Value"]),
            [
                ValidationError(
                    "Don't use 'NoEcho' parameter 'Echo' in 'Outputs'",
                    rule=NoEcho(),
                    path=deque(["Ref"]),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, path, expected, rule, context, cfn):
    context = context.evolve(path=Path(path=path))
    validator = CfnTemplateValidator(context=context, cfn=cfn)
    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules.jsonschema.MaxProperties import MaxProperties as ParentMaxProperties


class MaxProperties(ParentMaxProperties):
    """Check if maximum Resource limit is exceeded"""

    id = "E3010"
    shortdesc = "Resource limit not exceeded"
    description = (
        "Check the number of Resources in the template is less than the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["resources", "limits"]

    def __init__(self) -> None:
        super().__init__("I3010")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class Pattern(CloudFormationLintRule):
    """Check if maximum Parameter name size limit is exceeded"""

    id = "E2003"
    shortdesc = "Parameters have appropriate names"
    description = "Check if Parameters are properly named (A-Za-z0-9)"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["parameters", "name"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.jsonschema.PropertyNames import PropertyNames


class ChildMaxLength(CloudFormationLintRule):
    id = "ChildMaxLength"

    def foo(self, validator, awsType, instance, schema):
        if instance == "bar":
            return
        yield ValidationError("Not bar")


class TestMaxLength(BaseRuleTestCase):
    """Test AWS Types"""

    def setUp(self):
        """Setup"""
        super(TestMaxLength, self).setUp()
        self.rule = PropertyNames("ChildMaxLength")
        self.rule.child_rules = {"ChildMaxLength": ChildMaxLength()}

    def test_max_properties(self):
        validator = CfnTemplateValidator({})
        self.assertEqual(
            len(list(self.rule._max_length(validator, 5, "foo", {}))),
            0,
        )

        errs = list(self.rule._max_length(validator, 1, "foo", {}))
        self.assertEqual(len(errs), 1)
        self.assertIsNone(errs[0].rule)

        errs = list(self.rule._max_length(validator, 11, "1234567890", {}))
        self.assertEqual(len(errs), 1)
        self.assertEqual(errs[0].rule.id, "ChildMaxLength")

        errs = list(self.rule._max_length(validator, 11, {}, {}))
        self.assertListEqual(errs, [])


class ChildMaxLengthWithFunction(CloudFormationLintRule):
    id = "ChildMaxLengthWithFunction"

    def maxLength(self, validator, s, instance, schema):
        yield ValidationError("test", rule=self)
        yield ValidationError("test", rule=self)


class TestMaxLengthWithFunction(BaseRuleTestCase):
    """Test AWS Types"""

    def setUp(self):
        """Setup"""
        super(TestMaxLengthWithFunction, self).setUp()
        self.rule = PropertyNames("ChildMaxLengthWithFunction")
        self.rule.child_rules = {
            "ChildMaxLengthWithFunction": ChildMaxLengthWithFunction()
        }

    def test_property_names(self):
        validator = CfnTemplateValidator({})

        errs = list(self.rule._max_length(validator, 11, "1234567890", {}))
        self.assertEqual(len(errs), 2)
        self.assertEqual(errs[0].rule.id, "ChildMaxLengthWithFunction")


class TestMaxLengthNoChild(BaseRuleTestCase):
    """Test AWS Types"""

    def setUp(self):
        """Setup"""
        super(TestMaxLengthNoChild, self).setUp()
        self.rule = PropertyNames("ChildMaxLength")
        self.rule.child_rules = {"ChildMaxLength": None}

    def test_property_names(self):
        validator = CfnTemplateValidator({})

        errs = list(self.rule._max_length(validator, 11, "1234567890", {}))
        self.assertEqual(len(errs), 0)


class TestPropertyNames(BaseRuleTestCase):

    def test_rule_no_rule_id(self):
        validator = CfnTemplateValidator({})

        class Rule(PropertyNames):
            id = "AAAAA"

            def __init__(self):
                super(Rule, self).__init__("BBBBBB")

        rule = Rule()

        errs = list(
            rule.propertyNames(validator, {"enum": ["Bar"]}, {"Foo": "Bar"}, {})
        )

        self.assertEqual(len(errs), 1)
        self.assertIsNone(errs[0].rule)

    def test_rule_not_object(self):
        validator = CfnTemplateValidator({})

        class Rule(PropertyNames):
            id = "AAAAA"

            def __init__(self):
                super(Rule, self).__init__("BBBBBB")

        rule = Rule()

        errs = list(rule.propertyNames(validator, {"enum": ["Bar"]}, ["Bar"], {}))

        self.assertListEqual(errs, [])



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.schemas.other.resources
import cfnlint.helpers
from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails
from cfnlint.schema.resolver import RefResolver


class Configuration(CfnLintJsonSchema):
    """Check Update Policy Configuration"""

    id = "E3016"
    shortdesc = "Check the configuration of a resources UpdatePolicy"
    description = "Make sure a resources UpdatePolicy is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-updatepolicy.html"
    tags = ["resources", "updatepolicy"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/*"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.other.resources,
                filename="update_policy.json",
            ),
            all_matches=True,
        )

    def validate(self, validator: Validator, keywords: Any, instance: Any, schema: Any):
        validator = validator.evolve(
            context=validator.context.evolve(
                functions=list(FUNCTIONS),
                strict_types=False,
            ),
            schema=self._schema,
            resolver=RefResolver.from_schema(
                self._schema,
            ),
        )

        yield from super()._iter_errors(validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
from test.testlib.testcase import BaseTestCase
from unittest.mock import mock_open, patch

import cfnlint.helpers


class TestDownloadsMetadata(BaseTestCase):
    """Test Downloads Metadata"""

    @patch("cfnlint.helpers.os.path.exists")
    def test_no_file(self, mock_path_exists):
        """Test success run"""

        mock_path_exists.return_value = False

        filename = "foo.bar"

        result = cfnlint.helpers.load_metadata(filename)

        self.assertEqual(result, {})

    @patch("cfnlint.helpers.os.path.exists")
    def test_load_metadata(self, mock_path_exists):
        """Test success run"""

        mock_path_exists.return_value = True

        filename = "foo.bar"
        file_contents = {"etag": "foo"}

        builtin_module_name = "builtins"

        mo = mock_open(read_data=json.dumps(file_contents))
        with patch("{}.open".format(builtin_module_name), mo):
            result = cfnlint.helpers.load_metadata(filename)

            self.assertEqual(result, file_contents)

    @patch("cfnlint.helpers.os.path.exists")
    @patch("cfnlint.helpers.os.path.dirname")
    @patch("cfnlint.helpers.os.mkdir")
    @patch("cfnlint.helpers.json.dump")
    def test_save_download_metadata(
        self, mock_json_dump, mock_mkdir, mock_dirname, mock_path_exists
    ):
        """Test success run"""

        filename = "foo.bar"
        filedir = "foobardir"
        file_contents = {"etag": "foo"}

        mock_path_exists.return_value = False
        mock_dirname.return_value = filedir

        builtin_module_name = "builtins"

        mo = mock_open()
        with patch("{}.open".format(builtin_module_name), mo):
            cfnlint.helpers.save_metadata(file_contents, filename)
            mock_mkdir.assert_called_with(filedir)
            mock_json_dump.assert_called_once



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.rules.parameters.NumberRange import NumberRange as ParamaterNumberRange
from cfnlint.rules.resources.properties.NumberRange import NumberRange


@pytest.fixture
def rule():
    rule = NumberRange()
    rule.child_rules["W3034"] = ParamaterNumberRange()
    yield NumberRange()


def test_minimum(rule, validator):
    assert len(list(rule.minimum(validator, 1, 1, {}))) == 0
    assert len(list(rule.minimum(validator, 1, 0, {}))) == 1

    evolved = validator.evolve(
        context=validator.context.evolve(
            path=Path(
                path=deque(["Ref"]),
                value_path=deque(["Parameters", "MyParameter", "MinValue"]),
            )
        )
    )
    errs = list(rule.minimum(evolved, 1, 0, {}))
    assert len(errs) == 1
    assert errs[0].rule.id == ParamaterNumberRange.id

    rule.child_rules["W3034"] = None
    errs = list(rule.minimum(evolved, 1, 0, {}))
    assert len(errs) == 0


def test_maximum(rule, validator):
    assert len(list(rule.maximum(validator, 1, 1, {}))) == 0
    assert len(list(rule.maximum(validator, 1, 2, {}))) == 1

    evolved = validator.evolve(
        context=validator.context.evolve(
            path=Path(
                path=deque(["Ref"]),
                value_path=deque(["Parameters", "MyParameter", "MinValue"]),
            ),
        )
    )
    errs = list(rule.maximum(evolved, 1, 2, {}))
    assert len(errs) == 1
    assert errs[0].rule.id == ParamaterNumberRange.id

    rule.child_rules["W3034"] = None
    errs = list(rule.maximum(evolved, 1, 2, {}))
    assert len(errs) == 0


def test_exclusive_minimum(rule, validator):
    assert len(list(rule.exclusiveMinimum(validator, 1, 2, {}))) == 0
    assert len(list(rule.exclusiveMinimum(validator, 1, 1, {}))) == 1

    evolved = validator.evolve(
        context=validator.context.evolve(
            path=Path(
                path=deque(["Ref"]),
                value_path=deque(["Parameters", "MyParameter", "MinValue"]),
            ),
        )
    )
    errs = list(rule.exclusiveMinimum(evolved, 1, 1, {}))
    assert len(errs) == 1
    assert errs[0].rule.id == ParamaterNumberRange.id

    rule.child_rules["W3034"] = None
    errs = list(rule.exclusiveMinimum(evolved, 1, 1, {}))
    assert len(errs) == 0


def test_exlusive_maximum(rule, validator):
    assert len(list(rule.exclusiveMaximum(validator, 1, 0, {}))) == 0
    assert len(list(rule.exclusiveMaximum(validator, 1, 1, {}))) == 1

    evolved = validator.evolve(
        context=validator.context.evolve(
            path=Path(
                path=deque(["Ref"]),
                value_path=deque(["Parameters", "MyParameter", "MinValue"]),
            ),
        )
    )
    errs = list(rule.exclusiveMaximum(evolved, 1, 1, {}))
    assert len(errs) == 1
    assert errs[0].rule.id == ParamaterNumberRange.id

    rule.child_rules["W3034"] = None
    errs = list(rule.exclusiveMaximum(evolved, 1, 1, {}))
    assert len(errs) == 0



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint.formatters.base import BaseFormatter


class ParseableFormatter(BaseFormatter):
    """Parseable Formatter"""

    def _format(self, match):
        """Format output"""
        formatstr = "{0}:{1}:{2}:{3}:{4}:{5}:{6}"
        return formatstr.format(
            match.filename,
            match.linenumber,
            match.columnnumber,
            match.linenumberend,
            match.columnnumberend,
            match.rule.id,
            re.sub(r"(\r*\n)+", " ", match.message),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.context.context import Transforms
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.functions.Ref import Ref
from cfnlint.rules.functions.ToJsonString import ToJsonString
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = ToJsonString()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {"Resources": {"MyResource": {"Type": "AWS::S3::Bucket"}}},
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,schema,context_evolve,expected",
    [
        (
            "Fn::ToJsonString is not supported",
            {"Fn::ToJsonString": []},
            {"type": "string"},
            {},
            [
                ValidationError(
                    (
                        "Fn::ToJsonString is not supported without "
                        "'AWS::LanguageExtensions' transform"
                    ),
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_tojsonstring",
                    rule=ToJsonString(),
                ),
            ],
        ),
        (
            "Fn::ToJsonString is invalid with wrong output type",
            {"Fn::ToJsonString": {"foo": "bar"}},
            {"type": "object"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [
                ValidationError(
                    "{'Fn::ToJsonString': {'foo': 'bar'}} is not of type 'object'",
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_tojsonstring",
                    rule=ToJsonString(),
                ),
            ],
        ),
        (
            "Fn::ToJsonString is invalid with an empty object",
            {"Fn::ToJsonString": {}},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [
                ValidationError(
                    "{} does not have enough properties",
                    path=deque(["Fn::ToJsonString"]),
                    schema_path=deque(["minProperties"]),
                    validator="fn_tojsonstring",
                    rule=ToJsonString(),
                ),
            ],
        ),
        (
            "Fn::ToJsonString is invalid with an empty array",
            {"Fn::ToJsonString": []},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [
                ValidationError(
                    "[] is too short (1)",
                    path=deque(["Fn::ToJsonString"]),
                    schema_path=deque(["minItems"]),
                    validator="fn_tojsonstring",
                    rule=ToJsonString(),
                ),
            ],
        ),
        (
            "Fn::ToJsonString is invalid with a ref to AWS::NotificationARNs",
            {"Fn::ToJsonString": {"Ref": "AWS::NotificationARNs"}},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [
                ValidationError(
                    (
                        "'AWS::NotificationARNs' is not one of "
                        "['MyResource', 'AWS::AccountId', "
                        "'AWS::NoValue', 'AWS::Partition', 'AWS::Region', "
                        "'AWS::StackId', 'AWS::StackName', 'AWS::URLSuffix']"
                    ),
                    path=deque(["Fn::ToJsonString", "Ref"]),
                    schema_path=deque(["ref", "enum"]),
                    validator="ref",
                    rule=ToJsonString(),
                ),
            ],
        ),
        (
            "Fn::ToJsonString is valid array with functions",
            {"Fn::ToJsonString": [{"Ref": "MyResource"}]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [],
        ),
        (
            "Fn::ToJsonString is valid object with functions",
            {"Fn::ToJsonString": {"Key": {"Ref": "MyResource"}}},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            [],
        ),
    ],
)
def test_validate(name, instance, schema, context_evolve, expected, rule, context, cfn):
    context = context.evolve(**context_evolve)
    validator = CfnTemplateValidator({}).extend(
        validators={
            "ref": Ref().ref,
        }
    )(context=context, cfn=cfn)
    errs = list(rule.fn_tojsonstring(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.lmbd.FunctionZipfileRuntimeEnum import (
    FunctionZipfileRuntimeEnum,
)


@pytest.fixture(scope="module")
def rule():
    rule = FunctionZipfileRuntimeEnum()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"Code": {"ZipFile": "foo"}, "Runtime": "python3.10"},
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {"Code": {"ZipFile": "foo"}, "Runtime": {"Ref": "Runtime"}},
            [],
        ),
        (
            {"Code": {"ZipFile": "foo"}, "Runtime": "bar"},
            [
                ValidationError(
                    "'bar' does not match '^(nodejs.*|python.*)$'",
                    rule=FunctionZipfileRuntimeEnum(),
                    path=deque(["Runtime"]),
                    validator="pattern",
                    schema_path=deque(["then", "properties", "Runtime", "pattern"]),
                )
            ],
        ),
        (
            {"Code": "../link/to/my/package.zip", "Runtime": "provided.al2023"},
            [],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema


class Condition(CfnLintJsonSchema):

    id = "E6005"
    shortdesc = "Validate the Output condition is valid"
    description = (
        "Check the condition of an output to make sure it exists inside the template"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs", "conditions"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Outputs/*/Condition"],
            all_matches=True,
        )

    def validate(self, validator: Validator, _, instance: Any, schema):
        if not validator.is_type(instance, "string"):
            return

        validator = self.extend_validator(
            validator=validator,
            schema={"enum": list(validator.context.conditions.conditions.keys())},
            context=validator.context.evolve(),
        )

        yield from self._iter_errors(validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from unittest import TestCase

import cfnlint.decode.cfn_yaml


class BaseTestCase(TestCase):
    """
    All test cases should inherit from this class as any common
    functionality that is added here will then be available to all
    subclasses. This facilitates the ability to update in one spot
    and allow all tests to get the update for easy maintenance.
    """

    def load_template(self, filename):
        """Return template"""
        return cfnlint.decode.cfn_yaml.load(filename)

    def assertEqualListOfDicts(self, a, b):
        """Compare two lists of dicts"""
        assert isinstance(a, list)
        assert isinstance(b, list)

        def key_func(d):
            """sort dict based on keys"""
            items = ((k, v if v is not None else "") for k, v in d.items())
            return sorted(items)

        self.assertEqual(sorted(a, key=key_func), sorted(b, key=key_func))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class UniqueItemsAllowed(CloudFormationLintRule):
    """Check if duplicates exist in a List"""

    id = "I3037"
    shortdesc = "Check if a list that allows duplicates has any duplicates"
    description = (
        "Certain lists support duplicate items."
        "Provide an alert when list of strings or numbers have repeats."
    )
    source_url = (
        "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/rules.md#rules-1"
    )
    tags = ["resources", "property", "list"]

    def __init__(self):
        super().__init__()
        self.exceptions = ["Command"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.lmbd.ZipPackageRequiredProperties import (
    ZipPackageRequiredProperties,
)


class TestZipPackageRequiredProperties(BaseRuleTestCase):
    """Test required properties"""

    def setUp(self):
        super(TestZipPackageRequiredProperties, self).setUp()
        self.collection.register(ZipPackageRequiredProperties())
        self.success_templates = [
            "test/fixtures/templates/good/resources/lambda/required_properties.yaml"
        ]

    def test_file_positive(self):
        self.helper_file_positive()

    def test_file_negative(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/lambda/required_properties.yaml",
            err_count=3,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.jsonschema._keywords import pattern
from cfnlint.rules import CloudFormationLintRule


class ValuePattern(CloudFormationLintRule):
    """Check if parameters have a valid value"""

    id = "W2031"
    shortdesc = "Check if parameters have a valid value based on an allowed pattern"
    description = (
        "Check if parameters have a valid value in a pattern. The Parameter's allowed"
        " pattern is based on the usages in property (Ref)"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#pattern"
    tags = ["parameters", "resources", "property", "pattern"]

    # This rule is triggered from the equivalent rule E3031
    # the values are fed from there and we adjust the error outputs
    # appropriately

    # pylint: disable=unused-argument, arguments-renamed
    def pattern(self, validator, patrn, instance, schema):
        for err in pattern(validator, patrn, instance, schema):
            err.rule = self
            err.path_override = validator.context.path.value_path
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import unittest
from collections import deque

from cfnlint.jsonschema.exceptions import (
    FormatError,
    UndefinedTypeCheck,
    ValidationError,
    best_match,
)
from cfnlint.rules import CloudFormationLintRule


class _RuleX(CloudFormationLintRule):
    id = "X0000"


class _RuleY(CloudFormationLintRule):
    id = "Y0000"


class TestExceptions(unittest.TestCase):
    def test_validation_error(self):
        err = ValidationError(
            message="Failure", path=deque(["foo"]), schema_path=deque(["bar"])
        )

        self.assertListEqual(list(err.absolute_path), ["foo"])

        self.assertListEqual(list(err.absolute_schema_path), ["bar"])

        self.assertEqual(err.json_path, "$.foo")

    def test_validation_parent_error(self):
        parent = ValidationError(
            message="Failure", path=deque(["foo"]), schema_path=deque(["foo"])
        )
        err = ValidationError(
            message="Failure",
            path=deque(["bar", 0]),
            schema_path=deque(["bar"]),
            parent=parent,
        )

        self.assertListEqual(list(err.absolute_path), ["foo", "bar", 0])

        self.assertListEqual(list(err.absolute_schema_path), ["foo", "bar"])

        self.assertEqual(err.json_path, "$.foo.bar[0]")

    def test_best_match(self):
        one = ValidationError(message="One", path=deque(["foo", "bar", "foo"]))
        two = ValidationError(message="Two", path=deque(["foo", "bar"]))

        self.assertEqual(best_match([one, two]).message, "Two")

        three = ValidationError(
            message="Three",
            path=deque(["foo"]),
            context=[
                ValidationError(message="Four", path=deque(["a", "b"])),
                ValidationError(message="Five", path=deque(["a", "b", "c", "d"])),
                ValidationError(message="Size", path=deque(["a"])),
            ],
        )
        self.assertEqual(best_match([one, two, three]).message, "Five")

        self.assertEqual(best_match([]), None)

    def test_undefined_type_check(self):
        err = UndefinedTypeCheck("foo")
        self.assertEqual(str(err), "Type 'foo' is unknown to this type checker")

    def test_format_error(self):
        err = FormatError("foo")
        self.assertEqual(str(err), "foo")

    def test_equals(self):
        self.assertEqual(
            ValidationError(message="Failure", path=deque(["foo"])),
            ValidationError(message="Failure", path=deque(["foo"])),
        )

        self.assertNotEqual(
            ValidationError(message="Failure", path=deque(["foo"])),
            ValidationError(message="Failure", path=deque(["foo"]), rule=_RuleX()),
        )
        self.assertNotEqual(
            ValidationError(message="Failure", path=deque(["foo"]), rule=_RuleX()),
            ValidationError(
                message="Failure",
                path=deque(["foo"]),
            ),
        )
        self.assertNotEqual(
            ValidationError(message="Failure", path=deque(["foo"]), rule=_RuleX()),
            ValidationError(
                message="Failure",
                path=deque(["foo"]),
                rule=_RuleY(),
            ),
        )
        self.assertNotEqual(
            ValidationError(message="Failure", path=deque(["foo"]), rule=_RuleX()),
            "test",
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.DependsOn import DependsOn  # noqa: E501


@pytest.fixture(scope="module")
def rule():
    rule = DependsOn()
    yield rule


@pytest.fixture
def template():
    return {
        "Conditions": {
            "IsUsEast1": {
                "Fn::Equals": [
                    {"Ref": "AWS::Region"},
                    "us-east-1",
                ]
            }
        },
        "Resources": {
            "ParentBucket": {
                "Type": "AWS::S3::Bucket",
            },
            "ChildBucket": {
                "Type": "AWS::S3::Bucket",
                "DependsOn": "ParentBucket",
            },
            "ParentBucketWithGoodCondition": {
                "Condition": "IsUsEast1",
                "Type": "AWS::S3::Bucket",
            },
            "ChildBucketWithGoodCondition": {
                "Condition": "IsUsEast1",
                "DependsOn": "ParentBucketWithGoodCondition",
            },
            "ParentBucketWithBadCondition": {
                "Condition": "IsUsEast1",
                "Type": "AWS::S3::Bucket",
            },
            "ChildBucketWithBadCondition": {
                "Type": "AWS::S3::Bucket",
                "DependsOn": "ParentBucketWithBadCondition",
            },
            "ToMySelf": {
                "Type": "AWS::S3::Bucket",
                "DependsOn": "ToMySelf",
            },
        },
    }


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid depends on",
            "ParentBucket",
            deque(["Resources", "ChildBucket", "DependsOn"]),
            [],
        ),
        (
            "Invalid depends on but low path",
            "Foo",
            deque([]),
            [
                ValidationError(
                    (
                        "'Foo' is not one of ['ParentBucket', "
                        "'ChildBucket', "
                        "'ParentBucketWithGoodCondition', "
                        "'ParentBucketWithBadCondition', "
                        "'ChildBucketWithBadCondition', 'ToMySelf']"
                    ),
                )
            ],
        ),
        (
            "Valid depends on with conditions",
            "ParentBucketWithGoodCondition",
            deque(["Resources", "ChildBucketWithGoodCondition", "DependsOn"]),
            [],
        ),
        (
            "Invalid type isn't validated by this rule",
            {"Ref": "MyResource"},
            deque(["Resources", "ChildBucket", "DependsOn"]),
            [],
        ),
        (
            "Invalid depends on",
            "Foo",
            deque(["Resources", "ChildBucket", "DependsOn"]),
            [
                ValidationError(
                    (
                        "'Foo' is not one of ['ParentBucket', "
                        "'ParentBucketWithGoodCondition', "
                        "'ParentBucketWithBadCondition', "
                        "'ChildBucketWithBadCondition', 'ToMySelf']"
                    ),
                )
            ],
        ),
        (
            "Invalid depends on with conditions",
            "ParentBucketWithBadCondition",
            deque(["Resources", "ChildBucketWithBadCondition", "DependsOn"]),
            [
                ValidationError(
                    (
                        "'ParentBucketWithBadCondition' will not exist when condition "
                        "'IsUsEast1' is False"
                    ),
                )
            ],
        ),
        (
            "Invalid depends on relying on itself",
            "ToMySelf",
            deque(["Resources", "ToMySelf", "DependsOn"]),
            [
                ValidationError(
                    (
                        "'ToMySelf' is not one of ['ParentBucket', 'ChildBucket', "
                        "'ParentBucketWithGoodCondition', "
                        "'ParentBucketWithBadCondition', 'ChildBucketWithBadCondition']"
                    ),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, path, expected, rule, validator):

    validator = validator.evolve(context=validator.context.evolve(path=Path(path)))
    errs = list(rule.validate(validator, False, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"


def test_validate_with_no_cfn(rule, validator):
    validator = validator.evolve(
        context=validator.context.evolve(
            path=Path(path=deque(["Resources", "ChildBucket", "DependsOn"]))
        )
    )
    validator.cfn = None
    errs = list(rule.validate(validator, False, "ParentBucketWithBadCondition", {}))

    assert errs == [], f"Test without cfn got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_rds_dbcluster
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class DbClusterMonitoring(CfnLintJsonSchema):
    id = "E3689"
    shortdesc = "Validate MonitoringInterval and MonitoringRoleArn are used together"
    description = (
        "When MonitoringInterval is greater than 0 you need to specify "
        "MonitoringRoleArn. If MonitoringRoleArn is specified "
        "MonitoringInterval has to be greather than 0."
    )
    tags = ["resources"]
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-rds-dbcluster.html#cfn-rds-dbcluster-monitoringinterval"

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::RDS::DBCluster/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_rds_dbcluster,
                filename="monitoring.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return (
            "You must have 'MonitoringRoleArn' specified with "
            "'MonitoringInterval' greater than 0"
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

# ruff: noqa: E501
from cfnlint.rules.resources.PrimaryIdentifiers import PrimaryIdentifiers


class TestPrimaryIdentifiers(BaseRuleTestCase):
    """Test base template"""

    def setUp(self):
        """Setup"""
        super(TestPrimaryIdentifiers, self).setUp()
        self.collection.register(PrimaryIdentifiers())
        self.success_templates = [
            "test/fixtures/templates/good/resources/primary_identifiers.yaml",
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative_alias(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/primary_identifiers.yaml", 8
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.jsonschema import CfnLintKeyword


class DynamicReferenceSecret(CfnLintKeyword):
    """
    Check if Dynamic Reference Secure Strings are
    only used in the correct locations
    """

    id = "W1011"
    shortdesc = "Instead of REFing a parameter for a secret use a dynamic reference"
    description = (
        "Instead of REFing a parameter for a secret use a dynamic reference. "
        "Solutions like SSM parameter store and secrets manager provide "
        "better security of sercrets"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/security-best-practices.html#creds"
    tags = ["functions", "dynamic reference", "ref"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::DirectoryService::MicrosoftAD/Properties/Password",
                "Resources/AWS::DirectoryService::SimpleAD/Properties/Password",
                "Resources/AWS::ElastiCache::ReplicationGroup/Properties/AuthToken",
                "Resources/AWS::IAM::User/Properties/LoginProfile/Password",
                "Resources/AWS::KinesisFirehose::DeliveryStream/Properties/RedshiftDestinationConfiguration/Password",
                "Resources/AWS::OpsWorks::App/Properties/AppSource/Password",
                "Resources/AWS::OpsWorks::Stack/Properties/RdsDbInstances/*/DbPassword",
                "Resources/AWS::OpsWorks::Stack/Properties/CustomCookbooksSource/Password",
                "Resources/AWS::RDS::DBCluster/Properties/MasterUserPassword",
                "Resources/AWS::RDS::DBInstance/Properties/MasterUserPassword",
                "Resources/AWS::Redshift::Cluster/Properties/MasterUserPassword",
            ]
        )
        self.parent_rules = ["E1020"]

    def validate(self, validator: Validator, _, instance: Any, schema: Any):
        functions = set(FUNCTIONS) - set(["Fn::If"])
        if any(p in functions for p in validator.context.path.path):
            return
        value = instance.get("Ref")

        if not validator.is_type(value, "string"):
            return

        if value in validator.context.parameters:
            yield ValidationError(
                "Use dynamic references over parameters for secrets", rule=self
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
from typing import Any

import cfnlint.data.schemas.other.resources
import cfnlint.data.schemas.other.step_functions
import cfnlint.helpers
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails
from cfnlint.schema.resolver import RefResolver


class StateMachineDefinition(CfnLintJsonSchema):
    id = "E3601"
    shortdesc = "Validate the structure of a StateMachine definition"
    description = (
        "Validate the Definition or DefinitionString inside a "
        "AWS::StepFunctions::StateMachine resource"
    )
    source_url = "https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-state-machine-structure.html"
    tags = ["resources", "statemachine"]

    def __init__(self):
        super().__init__(
            keywords=[
                "Resources/AWS::StepFunctions::StateMachine/Properties/Definition",
                "Resources/AWS::StepFunctions::StateMachine/Properties/DefinitionString",
            ],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.step_functions, "statemachine.json"
            ),
            all_matches=True,
        )

        store = {
            "definition": self.schema,
        }

        self.resolver = RefResolver.from_schema(self.schema, store=store)

    def _fix_message(self, err: ValidationError) -> ValidationError:
        if len(err.path) > 1:
            err.message = f"{err.message} at {'/'.join(err.path)!r}"
        for i, c_err in enumerate(err.context):
            err.context[i] = self._fix_message(c_err)
        return err

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        # First time child rules are configured against the rule
        # so we can run this now
        add_path_to_message = False
        if validator.is_type(instance, "string"):
            try:
                step_validator = validator.evolve(
                    context=validator.context.evolve(
                        functions=[],
                    ),
                    resolver=self.resolver,
                    schema=self.schema,
                )
                instance = json.loads(instance)
                add_path_to_message = True
            except json.JSONDecodeError:
                return
        else:
            step_validator = validator.evolve(
                resolver=self.resolver,
                schema=self.schema,
            )

        for err in step_validator.iter_errors(instance):
            if add_path_to_message:
                err = self._fix_message(err)
            if not err.validator.startswith("fn_") and err.validator not in ["cfnLint"]:
                err.rule = self

            yield self._clean_error(err)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging

import pytest

from cfnlint.schema._schema import Schema
from cfnlint.schema.resolver._exceptions import RefResolutionError

LOGGER = logging.getLogger("cfnlint.schema.manager")
LOGGER.disabled = True


@pytest.fixture
def resource_schema():
    return {
        "additionalProperties": False,
        "definitions": {
            "ListItem": {"description": "A a list Id.", "type": "string"},
            "ResourceArn": {
                "description": "A resource ARN.",
                "maxLength": 256,
                "minLength": 1,
                "pattern": "^arn:aws.*$",
                "type": "string",
            },
            "AList": {
                "items": [
                    {"type": "string"},
                    {"type": "boolean"},
                ],
                "type": "array",
            },
        },
        "properties": {
            "Name": {
                "maxLength": 128,
                "minLength": 1,
                "pattern": "^[a-zA-Z0-9-]+$",
                "type": "string",
            },
            "List": {
                "insertionOrder": False,
                "items": {"$ref": "#/definitions/ListItem"},
                "type": "array",
            },
            "Arn": {"$ref": "#/definitions/ResourceArn"},
            "Id": {
                "maxLength": 36,
                "minLength": 36,
                "pattern": "^([0-9a-f]{8})-([0-9a-f]{4}-){3}([0-9a-f]{12})$",
                "type": "string",
            },
        },
        "readOnlyProperties": [
            "/properties/Arn",
            "/properties/Id",
            "/properties/List",
        ],
        "typeName": "AWS::NetworkFirewall::Firewall",
    }


def test_schema(resource_schema):

    schema = Schema(schema=resource_schema)

    assert schema.resolver.resolve_from_url("#/") == resource_schema

    assert schema.resolver.resolve_from_url("#/properties/Name") == {
        "maxLength": 128,
        "minLength": 1,
        "pattern": "^[a-zA-Z0-9-]+$",
        "type": "string",
    }

    assert schema.resolver.resolve_from_url("#/definitions/ListItem") == {
        "description": "A a list Id.",
        "type": "string",
    }
    assert schema.resolver.resolve_from_url("#/definitions/AList/items/0") == {
        "type": "string"
    }
    assert schema.resolver.resolve_from_url("#/definitions/AList/items/1") == {
        "type": "boolean"
    }

    with pytest.raises(RefResolutionError):
        schema.resolver.resolve_from_url("#/definitions/AList/items/2")

    with pytest.raises(RefResolutionError):
        schema.resolver.resolve_from_url("#/properties/bar/key")

    with pytest.raises(RefResolutionError):
        schema.resolver.resolve_from_url("test#/properties/bar/key")

    with pytest.raises(RefResolutionError):
        schema.resolver.pop_scope()
        schema.resolver.pop_scope()



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

import cfnlint.formatters
from cfnlint import ConfigMixIn
from cfnlint.runner import get_formatter


class TestGetFormatter(BaseTestCase):
    """Test get formatter with config"""

    def test_basic(self):
        config = ConfigMixIn([])
        formatter = get_formatter(config)

        self.assertIsInstance(formatter, cfnlint.formatters.Formatter)

    def test_json(self):
        config = ConfigMixIn(["--format", "json"])
        formatter = get_formatter(config)

        self.assertIsInstance(formatter, cfnlint.formatters.JsonFormatter)

    def test_quiet(self):
        config = ConfigMixIn(["--format", "quiet"])
        formatter = get_formatter(config)

        self.assertIsInstance(formatter, cfnlint.formatters.QuietFormatter)

    def test_parseable(self):
        config = ConfigMixIn(["--format", "parseable"])
        formatter = get_formatter(config)

        self.assertIsInstance(formatter, cfnlint.formatters.ParseableFormatter)

    def test_junit(self):
        config = ConfigMixIn(["--format", "junit"])
        formatter = get_formatter(config)

        self.assertIsInstance(formatter, cfnlint.formatters.JUnitFormatter)

    def test_pretty(self):
        config = ConfigMixIn(["--format", "pretty"])
        formatter = get_formatter(config)

        self.assertIsInstance(formatter, cfnlint.formatters.PrettyFormatter)

    def test_sarif(self):
        config = ConfigMixIn(["--format", "sarif"])
        formatter = get_formatter(config)

        self.assertIsInstance(formatter, cfnlint.formatters.SARIFFormatter)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import itertools
import operator

from cfnlint.formatters._utils import color, colored
from cfnlint.formatters.base import BaseFormatter


class PrettyFormatter(BaseFormatter):
    """Generic Formatter"""

    def _format(self, match):
        """Format output"""
        formatstr = "{0}{1}{2}"
        pos = f"{match.linenumber}:{match.columnnumber}:"
        return formatstr.format(
            colored(f"{pos:20}", color.reset),
            colored(f"{match.rule.id:10}", getattr(color, match.rule.severity.lower())),
            match.message,
        )

    def print_matches(self, matches, rules, config):
        results = self._format_matches(matches)

        # ruff: noqa: E501
        results.append(
            f"Cfn-lint scanned {colored(len(config.templates), color.bold_reset)} templates against "
            f"{colored(len(rules.used_rules), color.bold_reset)} rules and found "
            f'{colored(len([i for i in matches if i.rule.severity.lower() == "error"]), color.error)} '
            f'errors, {colored(len([i for i in matches if i.rule.severity.lower() == "warning"]), color.warning)} '
            f"warnings, and "
            f'{colored(len([i for i in matches if i.rule.severity.lower() == "informational"]), color.informational)} '
            f"informational violations"
        )
        return "\n".join(results)

    def _format_matches(self, matches):
        """Output all the matches"""
        output = []

        # This better be sorted
        for filename, file_matches in itertools.groupby(
            matches, key=operator.attrgetter("filename")
        ):
            levels = {"error": [], "warning": [], "informational": [], "unknown": []}

            output.append(colored(filename, color.underline_reset))
            for match in file_matches:
                level = match.rule.severity.lower()
                if level not in ["error", "warning", "informational"]:
                    level = "unknown"
                levels[level].append(match)
            for _, all_matches in levels.items():
                for match in all_matches:
                    output.extend([self._format(match)])

            output.append("")  # Newline after each group

        return output



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.metadata.InterfaceConfiguration import InterfaceConfiguration


@pytest.fixture(scope="module")
def rule():
    rule = InterfaceConfiguration()
    yield rule


@pytest.fixture
def template():
    return {
        "Parameters": {
            "Foo": {"Type": "String"},
            "Bar": {"Type": "String"},
        },
    }


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid Interface",
            {
                "ParameterGroups": [
                    {
                        "Label": "A Group",
                        "Parameters": [
                            "Foo",
                            "Bar",
                        ],
                    }
                ],
                "ParameterLabels": {"Foo": {"default": "a parameter"}},
            },
            [],
        ),
        (
            "Extra properties",
            {"Foo": "Bar"},
            [
                ValidationError(
                    ("Additional properties are not allowed " "('Foo' was unexpected)"),
                    validator="additionalProperties",
                    schema_path=deque(["additionalProperties"]),
                    rule=InterfaceConfiguration(),
                    path=deque(["Foo"]),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.helpers import REGIONS
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.GetAz import GetAz


@pytest.fixture(scope="module")
def rule():
    rule = GetAz()
    yield rule


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::GetAZs with empty string",
            {"Fn::GetAZs": ""},
            {"type": "array"},
            [],
        ),
        (
            "Valid Fn::GetAZs with Ref",
            {"Fn::GetAZs": {"Ref": "AWS::Region"}},
            {"type": "array"},
            [],
        ),
        (
            "Invalid Fn::GetAZs with an invalid output type",
            {"Fn::GetAZs": {"Ref": "AWS::Region"}},
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::GetAZs': {'Ref': 'AWS::Region'}} is not of type 'string'",
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_getazs",
                ),
            ],
        ),
        (
            "Invalid Fn::GetAZs with a bad value type",
            {"Fn::GetAZs": ["foo"]},
            {"type": "array"},
            [
                ValidationError(
                    "['foo'] is not of type 'string'",
                    path=deque(["Fn::GetAZs"]),
                    schema_path=deque(["type"]),
                    validator="fn_getazs",
                ),
                ValidationError(
                    f"['foo'] is not one of {[''] + REGIONS!r}",
                    path=deque(["Fn::GetAZs"]),
                    schema_path=deque(["enum"]),
                    validator="fn_getazs",
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.fn_getazs(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import gzip
from test.testlib.testcase import BaseTestCase
from unittest.mock import MagicMock, patch

import cfnlint.helpers


class TestGetUrlContent(BaseTestCase):
    """Test Get URL Content"""

    @patch("cfnlint.helpers.urlopen")
    def test_get_url_content_unzipped(self, mocked_urlopen):
        """Test success run"""

        input_buffer = '{"key": "value"}'

        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.read.return_value = input_buffer.encode("utf-8")
        cm.__enter__.return_value = cm
        mocked_urlopen.return_value = cm

        url = "http://foo.com"
        result = cfnlint.helpers.get_url_content(url)
        mocked_urlopen.assert_called_with(url)
        self.assertEqual(result, '{"key": "value"}')

    @patch("cfnlint.helpers.urlopen")
    def test_get_url_content_zipped(self, mocked_urlopen):
        """Test success run"""
        input_buffer = '{"key": "value"}'
        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.read.return_value = gzip.compress(input_buffer.encode("utf-8"))

        cm.info.return_value = {"Content-Encoding": "gzip"}
        cm.__enter__.return_value = cm
        mocked_urlopen.return_value = cm

        url = "http://foo.com"
        result = cfnlint.helpers.get_url_content(url)
        mocked_urlopen.assert_called_with(url)
        self.assertEqual(result, '{"key": "value"}')

    @patch("cfnlint.helpers.urlopen")
    @patch("cfnlint.helpers.load_metadata")
    @patch("cfnlint.helpers.save_metadata")
    def test_get_url_content_zipped_cache_update(
        self, mock_save_metadata, mock_load_metadata, mocked_urlopen
    ):
        """Test success run"""
        input_buffer = '{"key": "value"}'
        etag = "ETAG_ONE"
        url = "http://foo.com"

        mock_load_metadata.return_value = {}

        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.info.return_value = {"Content-Encoding": "gzip", "ETag": etag}

        cm.read.return_value = gzip.compress(input_buffer.encode("utf-8"))

        cm.__enter__.return_value = cm
        mocked_urlopen.return_value = cm

        result = cfnlint.helpers.get_url_content(url, caching=True)
        mocked_urlopen.assert_called_with(url)
        mock_load_metadata.assert_called_once()
        mock_save_metadata.assert_called_once()

        self.assertEqual(result, '{"key": "value"}')

    @patch("cfnlint.helpers.urlopen")
    @patch("cfnlint.helpers.load_metadata")
    def test_url_has_newer_version_affirmative(
        self, mock_load_metadata, mocked_urlopen
    ):
        """Test success run"""

        etag = "ETAG_ONE"
        url = "http://foo.com"

        mock_load_metadata.return_value = {"etag": etag}

        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.info.return_value = {"Content-Encoding": "gzip", "ETag": etag}

        cm.__enter__.return_value = cm
        mocked_urlopen.return_value = cm

        result = cfnlint.helpers.url_has_newer_version(url)

        # Python2 does not support caching, so will always return true
        self.assertFalse(result)

    @patch("cfnlint.helpers.urlopen")
    @patch("cfnlint.helpers.load_metadata")
    def test_url_has_newer_version_negative(self, mock_load_metadata, mocked_urlopen):
        """Test success run"""

        # Generate a random ETag to test with
        etag = "ETAG_ONE"
        etag2 = "ETAG_TWO"

        url = "http://foo.com"
        mock_load_metadata.return_value = {"etag": etag}

        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.info.return_value = {"Content-Encoding": "gzip", "ETag": etag2}
        cm.__enter__.return_value = cm
        mocked_urlopen.return_value = cm

        result = cfnlint.helpers.url_has_newer_version(url)
        self.assertTrue(result)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.jsonschema._keywords import format as parent_format
from cfnlint.rules import CloudFormationLintRule


class Format(CloudFormationLintRule):
    id = "E1103"
    shortdesc = "Validate the format of a value"
    description = "Parent rule for validating the format keyword in schemas"
    tags = []

    def format(
        self, validator: Validator, format: Any, instance: Any, schema: dict[str, Any]
    ):
        if validator.format_checker is not None:  # type: ignore
            if format in validator.format_checker.checkers:
                yield from parent_format(validator, format, instance, schema)
                return

        for rule in self.child_rules.values():
            if rule is None:
                continue
            if not rule.id:
                continue

            if format == rule.format_keyword:  # type: ignore
                if not rule.format(validator, instance):  # type: ignore
                    yield ValidationError(
                        f"{instance!r} is not a {format!r}",
                        rule=rule,
                    )
